/**
* blowfish.c
* this is part of semestral work to PROGRAMMING IN UNIX
* tomas.pop@seznam.cz
* this code is under Genereal public a General Public Documentation License 
*/
#include "blowfish.h"
#include "blowfish.2.h"	// holds the random digit tables
#include "endians.h"

#define S(bc,x,i) (bc->SBoxes[i][x.w.byte##i])
#define bf_F(bc,x) (((S(bc,x,0) + S(bc,x,1)) ^ S(bc,x,2)) + S(bc,x,3))
#define ROUND(bc, a,b,n) (a.dword ^= bf_F(bc,b) ^ bc->PArray[n])


void BlfInit(BlfCtx *bc, BYTE key[], int keybytes){
	int  		i, j ;
	DWORD  		data, datal, datar ;
	union aword temp ;

	// first fill arrays from data tables
	for (i = 0 ; i < 18 ; i++)
		bc->PArray [i] = bf_P [i] ;

	for (i = 0 ; i < 4 ; i++)
	{
	 	for (j = 0 ; j < 256 ; j++)
	 		bc->SBoxes [i][j] = bf_S [i][j] ;
	}


	j = 0 ;
	for (i = 0 ; i < NPASS + 2 ; ++i)
	{
		temp.dword = 0 ;
		temp.w.byte0 = key[j];
		temp.w.byte1 = key[(j+1) % keybytes] ;
		temp.w.byte2 = key[(j+2) % keybytes] ;
		temp.w.byte3 = key[(j+3) % keybytes] ;
		data = temp.dword ;
		bc->PArray [i] ^= data ;
		j = (j + 4) % keybytes ;
	}

	datal = 0 ;
	datar = 0 ;

	for (i = 0 ; i < NPASS + 2 ; i += 2)
	{
		BlfEncipher (bc,&datal, &datar) ;
		bc->PArray [i] = datal ;
		bc->PArray [i + 1] = datar ;
	}

	for (i = 0 ; i < 4 ; ++i)
	{
		for (j = 0 ; j < 256 ; j += 2)
		{
		  BlfEncipher (bc,&datal, &datar) ;
		  bc->SBoxes [i][j] = datal ;
		  bc->SBoxes [i][j + 1] = datar ;
		}
	}
	bc->cbc_ctx.lastL=0;
	bc->cbc_ctx.lastR=0;
}

void BlfEncipher(BlfCtx *bc, DWORD *xl, DWORD *xr){
	union aword  Xl, Xr ;

	Xl.dword = *xl ;
	Xr.dword = *xr ;

	Xl.dword ^= bc->PArray [0];
	ROUND (bc,Xr, Xl, 1) ;  ROUND (bc,Xl, Xr, 2) ;
	ROUND (bc,Xr, Xl, 3) ;  ROUND (bc,Xl, Xr, 4) ;
	ROUND (bc,Xr, Xl, 5) ;  ROUND (bc,Xl, Xr, 6) ;
	ROUND (bc,Xr, Xl, 7) ;  ROUND (bc,Xl, Xr, 8) ;
	ROUND (bc,Xr, Xl, 9) ;  ROUND (bc,Xl, Xr, 10) ;
	ROUND (bc,Xr, Xl, 11) ; ROUND (bc,Xl, Xr, 12) ;
	ROUND (bc,Xr, Xl, 13) ; ROUND (bc,Xl, Xr, 14) ;
	ROUND (bc,Xr, Xl, 15) ; ROUND (bc,Xl, Xr, 16) ;
	Xr.dword ^= bc->PArray [17] ;

	*xr = Xl.dword ;
	*xl = Xr.dword ;
}

void BlfDecipher(BlfCtx *bc, DWORD *xl, DWORD *xr){

   union aword  Xl ;
   union aword  Xr ;

   Xl.dword = *xl ;
   Xr.dword = *xr ;

   Xl.dword ^= bc->PArray [17] ;
   ROUND (bc,Xr, Xl, 16) ;  ROUND (bc,Xl, Xr, 15) ;
   ROUND (bc,Xr, Xl, 14) ;  ROUND (bc,Xl, Xr, 13) ;
   ROUND (bc,Xr, Xl, 12) ;  ROUND (bc,Xl, Xr, 11) ;
   ROUND (bc,Xr, Xl, 10) ;  ROUND (bc,Xl, Xr, 9) ;
   ROUND (bc,Xr, Xl, 8) ;   ROUND (bc,Xl, Xr, 7) ;
   ROUND (bc,Xr, Xl, 6) ;   ROUND (bc,Xl, Xr, 5) ;
   ROUND (bc,Xr, Xl, 4) ;   ROUND (bc,Xl, Xr, 3) ;
   ROUND (bc,Xr, Xl, 2) ;   ROUND (bc,Xl, Xr, 1) ;
   Xr.dword ^= bc->PArray[0];

   *xl = Xr.dword;
   *xr = Xl.dword;
}


DWORD GetOutputLength (DWORD lInputLong){
	DWORD 	lVal ;

	lVal = lInputLong % 8 ;	// find out if uneven number of bytes at the end
	if (lVal != 0)
		return lInputLong + 8 - lVal ;
	else
		return lInputLong ;
}


DWORD BlfEncode (BlfCtx *bc,BYTE * pInput, BYTE * pOutput, DWORD lSize){
	DWORD 	lCount, lOutSize, lGoodBytes ;
	BYTE	*pi, *po ;
	int		i, j ;
	int		SameDest = (pInput == pOutput ? 1 : 0) ;

	lOutSize = GetOutputLength (lSize) ;
	for (lCount = 0 ; lCount < lOutSize ; lCount += 8)
	{
		if (SameDest)	// if encoded data is being written into inputbuffer
		{
		 	if (lCount < lSize - 7)	// if not dealing with uneven bytes at end
		 	{
		 	 	BlfEncipher (bc,(DWORD *) pInput,(DWORD *) (pInput + 4)) ;
		 	}
		 	else		// pad end of data with null bytes to complete encryption
		 	{
				po = pInput + lSize ;	// point at byte past the end of actual data
				j = (int) (lOutSize - lSize) ;	// number of bytes to set to null
				for (i = 0 ; i < j ; i++)
					*po++ = 0 ;
		 	 	BlfEncipher (bc,(DWORD *) pInput,(DWORD *) (pInput + 4)) ;
		 	}
		 	pInput += 8 ;
		}
		else 			// output buffer not equal to input buffer, so must copy
		{               // input to output buffer prior to encrypting
		 	if (lCount < lSize - 7)	// if not dealing with uneven bytes at end
		 	{
		 		pi = pInput ;
		 		po = pOutput ;
		 		for (i = 0 ; i < 8 ; i++)
			// copy bytes to output
		 			*po++ = *pi++ ;
		 	 	BlfEncipher (bc,(DWORD *) pOutput,(DWORD *) (pOutput + 4)) ;
		 	}
		 	else		// pad end of data with null bytes to complete encryption
		 	{
		 		lGoodBytes = lSize - lCount ;	// number of remaining data bytes
		 		po = pOutput ;
		 		for (i = 0 ; i < (int) lGoodBytes ; i++)
		 			*po++ = *pInput++ ;
		 		for (j = i ; j < 8 ; j++)
		 			*po++ = 0 ;
		 	 	BlfEncipher (bc,(DWORD *) pOutput,(DWORD *) (pOutput + 4)) ;
		 	}
		 	pInput += 8 ;
		 	pOutput += 8 ;
		}
	}
	return lOutSize ;
}

void BlfDecode (BlfCtx *bc,BYTE * pInput, BYTE * pOutput, DWORD lSize){
	DWORD 	lCount ;
	BYTE	*pi, *po ;
	int		i ;
	int		SameDest = (pInput == pOutput ? 1 : 0) ;

	for (lCount = 0 ; lCount < lSize ; lCount += 8)
	{
		if (SameDest)	// if encoded data is being written into input buffer
		{
	 	 	BlfDecipher( bc,(DWORD *) pInput, (DWORD *) (pInput + 4)) ;
		 	pInput += 8 ;
		}
		else 			// output buffer not equal to input buffer
		{               // so copy input to output before decoding
	 		pi = pInput ;
	 		po = pOutput ;
	 		for (i = 0 ; i < 8 ; i++)
	 			*po++ = *pi++ ;
	 	 	BlfDecipher (bc,(DWORD *) pOutput,(DWORD *) (pOutput + 4)) ;
		 	pInput += 8 ;
		 	pOutput += 8 ;
		}
	}

}
void Blowfish_Encode_Block(BlfCtx *ctx, BYTE *block, unsigned int length){
  byteReverse(block, length/4);
  BlfEncode ( ctx, block, block, length);
  byteReverse(block, length/4);
}

void Blowfish_Decode_Block(BlfCtx *ctx, BYTE *block, unsigned int length){
  byteReverse(block, length/4);
  BlfDecode (ctx,  block,  block, length);
  byteReverse(block, length/4);
}
/*
void Blowfish_Encode_Block(BlfCtx *ctx, DWORD *block, unsigned int length){
	while (length>0){
     block[0] ^= ctx->cbc_ctx.lastL;
 	 block[1] ^= ctx->cbc_ctx.lastR;
	 BlfEncipher(ctx,block,block+1);
     ctx->cbc_ctx.lastL = block[0];
	 ctx->cbc_ctx.lastR = block[1];
	 length -= 2*4; //there are 4 bytes in in ulong 
	 block += 2;
	}

}


void Blowfish_Decode_Block(BlfCtx *ctx, DWORD *block, unsigned int length){
	DWORD cbc_tmp_l, cbc_tmp_r;
	while (length>0){
     cbc_tmp_l = ctx->cbc_ctx.lastL;
	 cbc_tmp_r = ctx->cbc_ctx.lastR;
     ctx->cbc_ctx.lastL = block[0];
	 ctx->cbc_ctx.lastR = block[1];
	 BlfDecipher(ctx,block,block+1);
	 block[0] ^= cbc_tmp_l;
	 block[1] ^= cbc_tmp_r;
	 length -= 2*4; //there are 4 bytes in in ulong
	 block += 2;
	}
}

*/
